/*----------------------------   element.h     ---------------------------*/
/*      $Id:$                 */
#ifndef __elementH
#define __elementH
/*----------------------------   element.h     ---------------------------*/

#include "particle.h"
#include "boundaryline.h"
#include <iomanip>
#include <random>

namespace GasParticle
{
  class Zufallszahlengenerator
  {
  private:
    std::default_random_engine generator;

  public:
    // Startet die Zufallszahlensuche mit einem "zufaelligem" Wert
    void initsimple()
    {
      int X = clock() + 1;
      generator.seed(X);
    }
    void init(double s)
    {
      int X = s * 10000.0;
      generator.seed(X);
    }

    void init()
    {
      std::cout << "Initialisierung der Zufallszahlen" << std::endl
		<< "Beliebige Eingabe: ";
      std::string S;
      std::cin >> S;
      const char *s = S.c_str();
      int X = clock() + 1;
      for (unsigned int i = 0; s[i] != 0; ++i)
	X *= s[i];
      generator.seed(X);
    }

    // liefert Zufallszahl im definierten Bereich
    int operator()(int min, int max)
    {
      return std::uniform_int_distribution<int>{min, max}(generator);
    }
    // liefert Zufallszahl im definierten Bereich
    double randomdouble(double min, double max)
    {
      return std::uniform_real_distribution<double>{min, max}(generator);
    }
    double normal(double a, double b)
    {
      return std::normal_distribution<double>{a, b}(generator);
    }
  };

  // One element of the lattice
  class Element
  {

  public:
    // Global Data of the Element
    double X0, Y0, X1, Y1; // coordinates of the element
    double H;			   // size of the element

    std::vector<int> boundary_lines_index;	 // indices of the boundary lines within this element
    std::vector<int> boundary_circles_index; // indices of the boundary circles within this element

    //
    std::vector<Particle> _P; // vector of particles... change to Pointer!

  public:
    Element() : X0(0), Y0(0), H(0)
    {
    }

    void BasicInit(double x0, double y0, double h)
    {
      boundary_lines_index.clear();

      assert(h > 0);

      H = h;
      X0 = x0;
      X1 = X0 + H;
      Y0 = y0;
      Y1 = Y0 + H;

    }

    /// access
    double x0() const
    {
      return X0;
    }
    double x1() const
    {
      return X1;
    }
    double y0() const
    {
      return Y0;
    }
    double y1() const
    {
      return Y1;
    }

    ////////// Boundary
    void AddBoundaryLine(int i)
    {
      boundary_lines_index.push_back(i);
    }
    bool cut_by_line(const BoundaryLine &bl) const
    {
      // check if corners of lattice are on different side of line
      // points on the line act add to both
      double omega[4] = {bl.nx * (X0 - bl.x0) + bl.ny * (Y0 - bl.y0),
			 bl.nx * (X1 - bl.x0) + bl.ny * (Y0 - bl.y0),
			 bl.nx * (X0 - bl.x0) + bl.ny * (Y1 - bl.y0),
			 bl.nx * (X1 - bl.x0) + bl.ny * (Y1 - bl.y0)};
      int np = 0;
      int nn = 0;
      for (int i = 0; i < 4; ++i)
	{
	  if (omega[i] > 0)
	    ++np;
	  if (omega[i] < 0)
	    ++nn;
	}
      return !((np == 4) || (nn == 4));
    }

    ////////////////////////////////////////////////// Particles
    const Particle &particle(int i) const
    {
      assert(i < _P.size());
      return _P[i];
    }
    Particle &particle(int i)
    {
      assert(i < _P.size());
      return _P[i];
    }

 
    
    

    inline  std::array<double,9>  iceconcentration() const 
    {
      std::array<double,9> element_concentration({0,0,0,0,0,0,0,0,0});

      /*
      1 2 3;
      4 0 5;
      6 7 8;
      */	
      
      for (int i = 0; i < _P.size(); ++i)
        {
			    
          // threshold where the center of the particle is allowed to lie in order to be inside the cell
	        double threshold_x0=X0+_P[i].radius-1e-8;
	        double threshold_x1=X1-_P[i].radius+1e-8;
	        double threshold_y0=Y0+_P[i].radius-1e-8;
	        double threshold_y1=Y1-_P[i].radius+1e-8;


          // assume that particle is not larger than one cell
          if (_P[i].radius > 0) // particle is active
	    {
		
        if ( (_P[i].x >= threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y <= threshold_y1) )
		{ // particle is in cell (cell no. 0)
        	element_concentration[0] += _P[i].area()/H/H;
        
        } else if ((_P[i].x >= threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y > threshold_y1)) 
		{ // particle is in cell no. 2 (up)
			double h2 = _P[i].radius - (Y1 - _P[i].y); //_P[i].y - threshold_y1;
			double alpha2 = 2*acos(1-(h2/_P[i].radius));
			double area2 = 0.5*_P[i].radius*_P[i].radius*(alpha2-sin(alpha2));
			element_concentration[2] += area2/H/H;
			element_concentration[0] += _P[i].area()/H/H - area2/H/H;

		} else if ((_P[i].x < threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y <= threshold_y1)) 
		{ // particle is in cell no. 4 (left)
			double h4 = _P[i].radius - (_P[i].x - X0); //threshold_x0 - _P[i].x;
			double alpha4 = 2*acos(1-(h4/_P[i].radius));
			double area4 = 0.5*_P[i].radius*_P[i].radius*(alpha4-sin(alpha4));
			element_concentration[4] += area4/H/H;
			element_concentration[0] += _P[i].area()/H/H - area4/H/H;
		
		} else if ((_P[i].x >= threshold_x0) && (_P[i].x > threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y <= threshold_y1)) 
		{ // particle is in cell no. 5 (right)
			double h5 = _P[i].radius - (X1 - _P[i].x); //_P[i].x - threshold_x1;
			double alpha5 = 2*acos(1-(h5/_P[i].radius));
			double area5 = 0.5*_P[i].radius*_P[i].radius*(alpha5-sin(alpha5));
			element_concentration[5] += area5/H/H;
			element_concentration[0] += _P[i].area()/H/H - area5/H/H;
		
		} else if ((_P[i].x >= threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y < threshold_y0) && (_P[i].y <= threshold_y1)) 
		{ // particle is in cell no. 7 (down)
			double h7 = _P[i].radius - (_P[i].y - Y0); //threshold_y0 - _P[i].y;
			double alpha7 = 2*acos(1-(h7/_P[i].radius));
			double area7 = 0.5*_P[i].radius*_P[i].radius*(alpha7-sin(alpha7));
			element_concentration[7] += area7/H/H;
			element_concentration[0] += _P[i].area()/H/H - area7/H/H;
            assert(element_concentration[7] >= 0);
          
		} else if ( (_P[i].x < threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y > threshold_y1) ) 
		{ // particle is in cell no. 1 (upper-left)
			// particle is only in cell 2 and 4
			double h2 = _P[i].radius - (Y1 - _P[i].y); //_P[i].y - threshold_y1; //
			assert(h2 >= 0);
			double h4 = _P[i].radius - (_P[i].x - X0); //threshold_x0 - _P[i].x; //x
			assert(h4 >= 0);
			double alpha2 = 2*acos(1-(h2/_P[i].radius));
			double alpha4 = 2*acos(1-(h4/_P[i].radius));
			assert(alpha2 >= 0);
			assert(alpha4 >= 0);
			double s2 = 2*_P[i].radius*sin(alpha2/2);
			double s4 = 2*_P[i].radius*sin(alpha4/2);
			assert(s2 >= 0);
			assert(s4 >= 0);
			double area2 = 0.5*_P[i].radius*_P[i].radius*(alpha2-sin(alpha2));
			double area4 = 0.5*_P[i].radius*_P[i].radius*(alpha4-sin(alpha4));
			assert(area2 >= 0);
			assert(area4 >= 0);
			double area1;
			// particle is also in cell 1
            double b = s2/2 - (_P[i].x - X0);// X0 - (_P[i].x - s2/2);
            double a = s4/2 - (Y1 - _P[i].y); //(s4/2 + _P[i].y) - Y1;
            double striangle = sqrt(a*a + b*b);
            double area_triangle = 0.5*a*b;
            double h_c = 2*area_triangle/striangle;
            if (h_c <= 0 || a <= 0 || b <= 0)
            {
                area1 = 0;
            } else {
            double center_to_corner = sqrt((_P[i].x-X0)*(_P[i].x-X0) + (Y1 - _P[i].y)*(Y1 - _P[i].y));
			assert(center_to_corner < _P[i].radius);
            double h1 = _P[i].radius - (center_to_corner + h_c);
			assert(h_c >= 0);
			assert(h1 >= 0);
            double alpha1 = 2*acos(1-(h1/_P[i].radius));
			assert(alpha1 >= 0);
			assert(center_to_corner >= 0);
			assert(area_triangle >= 0);
            area1 = area_triangle + 0.5*_P[i].radius*_P[i].radius*(alpha1 - sin(alpha1));
            };
            element_concentration[1] += area1/H/H;
            
            element_concentration[0] += _P[i].area()/H/H;
			element_concentration[2] += (area2-area1)/H/H;
			element_concentration[4] += (area4-area1)/H/H;
			element_concentration[0] -= (area2/H/H + area4/H/H - area1/H/H); 
			
		} else if ( (_P[i].x >= threshold_x0) && (_P[i].x > threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y > threshold_y1) ) 
		{ // particle is in cell no. 3 (upper-right)
			// particle is only in cell 2 and 5
			double h2 = _P[i].radius - (Y1 - _P[i].y); //_P[i].y - threshold_y1; //y
			assert(h2>=0);
			double h5 = _P[i].radius - (X1 - _P[i].x); //_P[i].x - threshold_x1; //x
			assert(h5>=0);
			double alpha2 = 2*acos(1-(h2/_P[i].radius));
			double alpha5 = 2*acos(1-(h5/_P[i].radius));
			double s2 = 2*_P[i].radius*sin(alpha2/2);
			double s5 = 2*_P[i].radius*sin(alpha5/2);
			double area2 = 0.5*_P[i].radius*_P[i].radius*(alpha2-sin(alpha2));
			double area5 = 0.5*_P[i].radius*_P[i].radius*(alpha5-sin(alpha5));
			double area3;
			// particle is also in cell 3
            double b = s2/2 - (X1 - _P[i].x);//(s2/2 + _P[i].x) - X1;
            double a = s5/2 - (Y1 - _P[i].y);//(s5/2 + _P[i].x) - Y1;
            double s_triangle = sqrt(a*a + b*b);
            double area_triangle = 0.5*a*b;
            double h_c = 2*area_triangle/s_triangle;
            if (h_c <= 0 || a <= 0 || b <= 0)
            {
                area3 = 0;
            } else {
				double center_to_corner = sqrt((X1-_P[i].x)*(X1-_P[i].x) + (Y1 - _P[i].y)*(Y1 - _P[i].y));
				double h3 = _P[i].radius - (center_to_corner + h_c);
                double alpha3 = 2*acos(1-(h3/_P[i].radius));
                area3 = area_triangle + 0.5*_P[i].radius*_P[i].radius*(alpha3 - sin(alpha3));
            };
            element_concentration[3] += area3/H/H;

			element_concentration[0] += _P[i].area()/H/H;
            element_concentration[2] += (area2-area3)/H/H;
			element_concentration[5] += (area5-area3)/H/H;
			element_concentration[0] -= (area2/H/H + area5/H/H - area3/H/H); 
			
		} else if ( (_P[i].x < threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y < threshold_y0) && (_P[i].y <= threshold_y1) ) 
		{ // particle is in cell no. 6 (lower-left)
			// particle is only in cell 4 and 7
			double h4 = _P[i].radius - (_P[i].x - X0); //threshold_x0 - _P[i].x; //x
			assert(h4>=0);
			double h7 = _P[i].radius - (_P[i].y - Y0); //threshold_y0 - _P[i].y; //y
			assert(h7>=0);
			double alpha4 = 2*acos(1-(h4/_P[i].radius));
			double alpha7 = 2*acos(1-(h7/_P[i].radius));
            assert(alpha4>=0);
            assert(alpha7>=0);
			double s4 = 2*_P[i].radius*sin(alpha4/2);
			double s7 = 2*_P[i].radius*sin(alpha7/2);
            assert(s4>=0);
            assert(s7>=0);
			double area4 = 0.5*_P[i].radius*_P[i].radius*(alpha4-sin(alpha4));
			double area7 = 0.5*_P[i].radius*_P[i].radius*(alpha7-sin(alpha7));
			double area6;
			// particle is also in cell 6
            double b = s7/2 - (_P[i].x - X0);//Y0 - (_P[i].x - s7/2);
            double a = s4/2 - (_P[i].y - Y0);//X0 - (_P[i].y - s4/2);
            double s_triangle = sqrt(a*a + b*b);
            double area_triangle = 0.5*a*b;
            double h_c = 2*area_triangle/s_triangle;
          
			if (h_c <= 0 || a <= 0 || b <= 0)
            {
                area6 = 0;
            } else {
                double center_to_corner = sqrt((_P[i].x - X0)*(_P[i].x - X0) + (_P[i].y - Y0)*(_P[i].y - Y0));
                assert(center_to_corner < _P[i].radius);
				double h6 = _P[i].radius - (center_to_corner + h_c);
                double alpha6 = 2*acos(1-(h6/_P[i].radius));
                area6 = area_triangle + 0.5*_P[i].radius*_P[i].radius*(alpha6 - sin(alpha6));
            };
            element_concentration[6] += area6/H/H;

            element_concentration[0] += _P[i].area()/H/H;
			element_concentration[4] += (area4-area6)/H/H;
			element_concentration[7] += (area7-area6)/H/H;
			element_concentration[0] -= (area4/H/H + area7/H/H - area6/H/H); 
            assert(element_concentration[7] >= 0);

            double teil0 =  _P[i].area()/H/H - (area4/H/H + area7/H/H - area6/H/H);
			double test = teil0 + (area4-area6)/H/H + (area7-area6)/H/H + area6/H/H;
          
		} else if ( (_P[i].x >= threshold_x0) && (_P[i].x > threshold_x1) && (_P[i].y < threshold_y0) && (_P[i].y <= threshold_y1) ) 
		{ // particle is in cell no. 8 (lower-right)
			// particle is only in cell 5 and 7
			double h5 = _P[i].radius - (X1 - _P[i].x); //_P[i].x - threshold_x1; //x
			assert(h5>=0);
			double h7 = _P[i].radius - (_P[i].y - Y0); //threshold_y0 - _P[i].y; //y
			assert(h7>=0);
			double alpha5 = 2*acos(1-(h5/_P[i].radius));
            assert(alpha5>=0);
			double alpha7 = 2*acos(1-(h7/_P[i].radius));
            assert(alpha7>=0);
			double s5 = 2*_P[i].radius*sin(alpha5/2);
			double s7 = 2*_P[i].radius*sin(alpha7/2);
            assert(s5>=0);
            assert(s7>=0);
			double area5 = 0.5*_P[i].radius*_P[i].radius*(alpha5-sin(alpha5));
			double area7 = 0.5*_P[i].radius*_P[i].radius*(alpha7-sin(alpha7));
			double area8;
			// particle is also in cell 8
            double b = s5/2 - (_P[i].y - Y0);//Y0 - (_P[i].x - s5/2);
            double a = s7/2 - (X1 - _P[i].x);//(_P[i].x + s7/2) - X1;
            double s_triangle = sqrt(a*a + b*b);
            double area_triangle = 0.5*a*b;
            double h_c = 2*area_triangle/s_triangle;
            if (h_c <= 0 || a <= 0 || b <= 0)
            {
				area8 = 0;
            } else {
                double center_to_corner = sqrt((X1 - _P[i].x)*(X1 - _P[i].x) + (_P[i].y - Y0)*(_P[i].y - Y0));
				double h8 = _P[i].radius - (center_to_corner + h_c);
                double alpha8 = 2*acos(1-(h8/_P[i].radius));
                area8 = area_triangle + 0.5*_P[i].radius*_P[i].radius*(alpha8 - sin(alpha8));
            };
           
            element_concentration[8] += area8/H/H;
            
            element_concentration[0] += _P[i].area()/H/H;
			element_concentration[5] += (area5-area8)/H/H;
			element_concentration[7] += (area7-area8)/H/H;
			element_concentration[0] -= (area5/H/H + area7/H/H - area8/H/H);
		
		} else
        {
            std::cout << "element.h: no matching if case" << std::endl;
            abort();
        };
        
    
	    } // if particle is active
	
	  
	} // big for loop
       assert(element_concentration[0] >= 0);
	   assert(element_concentration[1] >= 0);
	   assert(element_concentration[2] >= 0);
	   assert(element_concentration[3] >= 0);
	   assert(element_concentration[4] >= 0);
	   assert(element_concentration[5] >= 0);
	   assert(element_concentration[6] >= 0);
	   assert(element_concentration[7] >= 0);
	   assert(element_concentration[8] >= 0);
       return element_concentration;
    } // inline	
	

    inline  std::array<double,9>  icevolume() const 
    {
      std::array<double,9> element_volume({0,0,0,0,0,0,0,0,0});

      /*
      1 2 3;
      4 0 5;
      6 7 8;
      */	
      
      for (int i = 0; i < _P.size(); ++i)
        {
			    
          // threshold where the center of the particle is allowed to lie in order to be inside the cell
	        double threshold_x0=X0+_P[i].radius-1e-8;
	        double threshold_x1=X1-_P[i].radius+1e-8;
	        double threshold_y0=Y0+_P[i].radius-1e-8;
	        double threshold_y1=Y1-_P[i].radius+1e-8;


          // assume that particle is not larger than one cell
          if (_P[i].radius > 0) // particle is active
	    {


        // particle is in cell (cell no. 0)
        if ( (_P[i].x >= threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y <= threshold_y1) ){
            element_volume[0] += _P[i].area()*_P[i].height;
            
        } else

              // particle is in cell no. 2 (up)
			  if ((_P[i].x >= threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y > threshold_y1)) {
			  	double h2 = _P[i].radius - (Y1 - _P[i].y); //_P[i].y - threshold_y1;
			  	double alpha2 = 2*acos(1-(h2/_P[i].radius));
			  	//double s2 = 2*_P[i].radius * sin(alpha2/2);
			  	double area2 = 0.5*_P[i].radius*_P[i].radius*(alpha2-sin(alpha2));
			  	element_volume[2] += area2*_P[i].height;
			  	element_volume[0] += _P[i].area()*_P[i].height - area2*_P[i].height;
			  
			  } else
			  // particle is in cell no. 4 (left)
			  if ((_P[i].x < threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y <= threshold_y1)) {
			  	double h4 = _P[i].radius - (_P[i].x - X0); //threshold_x0 - _P[i].x;
			  	double alpha4 = 2*acos(1-(h4/_P[i].radius));
			  	//double s4 = 2*_P[i].radius*sin(alpha4/2);
			  	double area4 = 0.5*_P[i].radius*_P[i].radius*(alpha4-sin(alpha4));
			  	element_volume[4] += area4*_P[i].height;
			  	element_volume[0] += _P[i].area()*_P[i].height - area4*_P[i].height;
			  	
			  } else
			  // particle is in cell no. 5 (right)
			  if ((_P[i].x >= threshold_x0) && (_P[i].x > threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y <= threshold_y1)) {
			  	double h5 = _P[i].radius - (X1 - _P[i].x); //_P[i].x - threshold_x1;
			  	double alpha5 = 2*acos(1-(h5/_P[i].radius));
			  	//double s5 = 2*_P[i].radius*sin(alpha5/2);
			  	double area5 = 0.5*_P[i].radius*_P[i].radius*(alpha5-sin(alpha5));
			  	element_volume[5] += area5*_P[i].height;
			  	element_volume[0] += _P[i].area()*_P[i].height - area5*_P[i].height;
			  	
			  } else
			  // particle is in cell no. 7 (down)
			  if ((_P[i].x >= threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y < threshold_y0) && (_P[i].y <= threshold_y1)) {
				  double h7 = _P[i].radius - (_P[i].y - Y0); //threshold_y0 - _P[i].y;
				  double alpha7 = 2*acos(1-(h7/_P[i].radius));
				  //double s7 = 2*_P[i].radius*sin(alpha7/2);
				  double area7 = 0.5*_P[i].radius*_P[i].radius*(alpha7-sin(alpha7));
				  element_volume[7] += area7*_P[i].height;
				  element_volume[0] += _P[i].area()*_P[i].height - area7*_P[i].height;
				  
			  } else if ( (_P[i].x < threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y > threshold_y1) ) 
		{ // particle is in cell no. 1 (upper-left)
			// particle is only in cell 2 and 4
			double h2 = _P[i].radius - (Y1 - _P[i].y); //_P[i].y - threshold_y1; //y
			double h4 = _P[i].radius - (_P[i].x - X0); //threshold_x0 - _P[i].x; //x
			double alpha2 = 2*acos(1-(h2/_P[i].radius));
			double alpha4 = 2*acos(1-(h4/_P[i].radius));
			double s2 = 2*_P[i].radius*sin(alpha2/2);
			double s4 = 2*_P[i].radius*sin(alpha4/2);
			double area2 = 0.5*_P[i].radius*_P[i].radius*(alpha2-sin(alpha2));
			double area4 = 0.5*_P[i].radius*_P[i].radius*(alpha4-sin(alpha4));
			double area1;
			// particle is also in cell 1
			double b = s2/2 - (_P[i].x - X0);// X0 - (_P[i].x - s2/2);
            double a = s4/2 - (Y1 - _P[i].y); //(s4/2 + _P[i].y) - Y1;
            double s_triangle = sqrt(a*a + b*b);
            double area_triangle = 0.5*a*b;
            double h_c = 2*area_triangle/s_triangle;
            if (h_c <= 0 || a <= 0 || b <= 0)
            {
                area1 = 0;
            } else {
				double center_to_corner = sqrt((_P[i].x-X0)*(_P[i].x-X0) + (Y1 - _P[i].y)*(Y1 - _P[i].y));
				double h1 = _P[i].radius - (center_to_corner + h_c);
				double alpha1 = 2*acos(1-(h1/_P[i].radius));
				area1 = area_triangle + 0.5*_P[i].radius*_P[i].radius*(alpha1 - sin(alpha1));
            };
            element_volume[1] += area1*_P[i].height;
			
            element_volume[0] += _P[i].area()*_P[i].height;
			element_volume[2] += (area2-area1)*_P[i].height;
			element_volume[4] += (area4-area1)*_P[i].height;
			element_volume[0] -= (area2*_P[i].height + area4*_P[i].height - area1*_P[i].height); // Saskia
           
		} else if ( (_P[i].x >= threshold_x0) && (_P[i].x > threshold_x1) && (_P[i].y >= threshold_y0) && (_P[i].y > threshold_y1) ) 
		{ // particle is in cell no. 3 (upper-right)
			// particle is only in cell 2 and 5
			double h2 = _P[i].radius - (Y1 - _P[i].y); //_P[i].y - threshold_y1; //y
			double h5 = _P[i].radius - (X1 - _P[i].x); //_P[i].x - threshold_x1; //x
			double alpha2 = 2*acos(1-(h2/_P[i].radius));
			double alpha5 = 2*acos(1-(h5/_P[i].radius));
			double s2 = 2*_P[i].radius*sin(alpha2/2);
			double s5 = 2*_P[i].radius*sin(alpha5/2);
			double area2 = 0.5*_P[i].radius*_P[i].radius*(alpha2-sin(alpha2));
			double area5 = 0.5*_P[i].radius*_P[i].radius*(alpha5-sin(alpha5));
			double area3;
			// particle is also in cell 3
            double b = s2/2 - (X1 - _P[i].x);//(s2/2 + _P[i].x) - X1;
            double a = s5/2 - (Y1 - _P[i].y);//(s5/2 + _P[i].x) - Y1;
            double s_triangle = sqrt(a*a + b*b);
            double area_triangle = 0.5*a*b;
            double h_c = 2*area_triangle/s_triangle;
			if (h_c <= 0 || a <= 0 || b <= 0)
            {
                area3 = 0;
            } else {
                double center_to_corner = sqrt((X1-_P[i].x)*(X1-_P[i].x) + (Y1 - _P[i].y)*(Y1 - _P[i].y));
				double h3 = _P[i].radius - (center_to_corner + h_c);
                double alpha3 = 2*acos(1-(h3/_P[i].radius));
                area3 = area_triangle + 0.5*_P[i].radius*_P[i].radius*(alpha3 - sin(alpha3));
            }
            element_volume[3] += area3*_P[i].height;
            
            element_volume[0] += _P[i].area()*_P[i].height;
			element_volume[2] += (area2-area3)*_P[i].height;
			element_volume[5] += (area5-area3)*_P[i].height;
			element_volume[0] -= (area2*_P[i].height + area5*_P[i].height - area3*_P[i].height); // Saskia
			
		} else if ( (_P[i].x < threshold_x0) && (_P[i].x <= threshold_x1) && (_P[i].y < threshold_y0) && (_P[i].y <= threshold_y1) ) 
		{ // particle is in cell no. 6 (lower-left)
			// particle is only in cell 4 and 7
			double h4 = _P[i].radius - (_P[i].x - X0); //threshold_x0 - _P[i].x; //x
			double h7 = _P[i].radius - (_P[i].y - Y0); //threshold_y0 - _P[i].y; //y
			double alpha4 = 2*acos(1-(h4/_P[i].radius));
			double alpha7 = 2*acos(1-(h7/_P[i].radius));
			double s4 = 2*_P[i].radius*sin(alpha4/2);
			double s7 = 2*_P[i].radius*sin(alpha7/2);
			double area4 = 0.5*_P[i].radius*_P[i].radius*(alpha4-sin(alpha4));
			double area7 = 0.5*_P[i].radius*_P[i].radius*(alpha7-sin(alpha7));
			double area6;
			// particle is also in cell 6
            double b = s7/2 - (_P[i].x - X0);//Y0 - (_P[i].x - s7/2);
            double a = s4/2 - (_P[i].y - Y0);//X0 - (_P[i].y - s4/2);
            double s_triangle = sqrt(a*a + b*b);
            double area_triangle = 0.5*a*b;
            double h_c = 2*area_triangle/s_triangle;
            if (h_c <= 0 || a <= 0 || b <= 0)
            {
                area6 = 0;
            } else {
				double center_to_corner = sqrt((_P[i].x - X0)*(_P[i].x - X0) + (_P[i].y - Y0)*(_P[i].y - Y0));
				double h6 = _P[i].radius - (center_to_corner + h_c);
                double alpha6 = 2*acos(1-(h6/_P[i].radius));
                area6 = area_triangle + 0.5*_P[i].radius*_P[i].radius*(alpha6 - sin(alpha6));
            };
            element_volume[6] += area6*_P[i].height;
			
            element_volume[0] += _P[i].area()*_P[i].height;
			element_volume[4] += (area4-area6)*_P[i].height;
			element_volume[7] += (area7-area6)*_P[i].height;
			element_volume[0] -= (area4*_P[i].height + area7*_P[i].height - area6*_P[i].height); // Saskia
		
		} else if ( (_P[i].x >= threshold_x0) && (_P[i].x > threshold_x1) && (_P[i].y < threshold_y0) && (_P[i].y <= threshold_y1) ) 
		{ // particle is in cell no. 8 (lower-right)
			// particle is only in cell 5 and 7
			double h5 = _P[i].radius - (X1 - _P[i].x); //_P[i].x - threshold_x1; //x
			double h7 = _P[i].radius - (_P[i].y - Y0); //threshold_y0 - _P[i].y; //y
			double alpha5 = 2*acos(1-(h5/_P[i].radius));
			double alpha7 = 2*acos(1-(h7/_P[i].radius));
			double s5 = 2*_P[i].radius*sin(alpha5/2);
			double s7 = 2*_P[i].radius*sin(alpha7/2);
			double area5 = 0.5*_P[i].radius*_P[i].radius*(alpha5-sin(alpha5));
			double area7 = 0.5*_P[i].radius*_P[i].radius*(alpha7-sin(alpha7));
			double area8;
			// particle is also in cell 8
            double b = s5/2 - (_P[i].y - Y0);//Y0 - (_P[i].x - s5/2);
            double a = s7/2 - (X1 - _P[i].x);//(_P[i].x + s7/2) - X1;
            double s_triangle = sqrt(a*a + b*b);
            double area_triangle = 0.5*a*b;
            double h_c = 2*area_triangle/s_triangle;
            if (h_c <= 0 || a <= 0 || b <= 0)
            {
                area8 = 0;
            } else {
				double center_to_corner = sqrt((X1 - _P[i].x)*(X1 - _P[i].x) + (_P[i].y - Y0)*(_P[i].y - Y0));
				double h8 = _P[i].radius - (center_to_corner + h_c);
                double alpha8 = 2*acos(1-(h8/_P[i].radius));
                area8 = area_triangle + 0.5*_P[i].radius*_P[i].radius*(alpha8 - sin(alpha8));
            };
            element_volume[8] += area8*_P[i].height;

            element_volume[0] += _P[i].area()*_P[i].height;
			element_volume[5] += (area5-area8)*_P[i].height;
			element_volume[7] += (area7-area8)*_P[i].height;
			element_volume[0] -= (area5*_P[i].height + area7*_P[i].height - area8*_P[i].height); // Saskia
			
		} else
        {
            std::cout << "element.h: no matching if case" << std::endl;
            abort();
        };
        
    
	    } // if particle is active
	
	  
	} // big for loop
       
       return element_volume;
    } // inline	
    
    void AddParticle(const Particle &p)
    {
      _P.push_back(p);
    }

    int nparticles() const
    {
      return _P.size();
    }

    void clear_deleted_particles()
    {
      int np = 0; // index for next particle
      for (int i = 0; i < _P.size(); ++i)
	{
	  if (_P[i].radius > 0) // particle is active
	    {
	      if (np < i) // must copy?
		_P[np] = _P[i];
	      ++np;
	    }
	}
      _P.resize(np); // shorten particle vector
    }

    // marks all particles outside the domain for deletion
    void delete_outside_particles(const std::vector<BoundaryLine> &BL, const std::vector<BoundaryCircle> &BC)
    {
      for (const auto it : boundary_lines_index) // lines
	for (int i = 0; i < _P.size(); ++i)
	  {
	    if (BL[it].distance(_P[i]) > 0)
	      _P[i].radius = -1;
	  }
      for (const auto it : boundary_circles_index) // circles
	for (int i = 0; i < _P.size(); ++i)
	  {
	    if (BC[it].distance(_P[i]) > 0)
	      _P[i].radius = -1;
	  }
      // std::cout << "remove!" << std::endl;
      for (int i = 0; i < _P.size(); ++i)
	{
	  if (pow(_P[i].x - 5., 2.0) + pow(_P[i].y + 0.75, 2.0) < 1.25 * 1.25)
	    _P[i].radius = -1;
	}
    }

    // marks all particles that overlap & and move towards each other
    void delete_overlapping_particles()
    {
      for (int i = 0; i < _P.size();++i)
	for (int j = i + 1; j < _P.size(); ++j)
	  {
	    if (_P[i].outer_distance(_P[j]) < 0)
	      {
		// get relative position and velocity
		double nx = _P[i].x - _P[j].x;
		double ny = _P[i].y - _P[j].y;
		double vx = _P[i].vx - _P[j].vx;
		double vy = _P[i].vy - _P[j].vy;
		if (nx * vx + ny * vy < 0) // delete particle if move towards each other
		  _P[i].radius = -1;
	      }
	  }
    }
    void delete_overlapping_particles(const Element &el)
    {
      for (int i = 0; i < _P.size(); ++i)
	for (int j = 0; j < el._P.size(); ++j)
	  {
	    if (_P[i].outer_distance(el._P[j]) < 0)
	      {
		// get relative position and velocity
		double nx = _P[i].x - el._P[j].x;
		double ny = _P[i].y - el._P[j].y;
		double vx = _P[i].vx - el._P[j].vx;
		double vy = _P[i].vy - el._P[j].vy;
		if (nx * vx + ny * vy < 0) // delete particle if move towards each other
		  _P[i].radius = -1;
	      }
	  }
    }

    // maximum velocity of the particles
    double vmax() const
    {
      double vm = 0;
      for (auto p : _P)
	if (p.radius > 0)
	  vm = std::max(vm, p.vnorm());
      return vm;
    }

    void InitRandom(int n,
		    double minradius, double maxradius,
		    double density,
		    const std::vector<BoundaryLine> &BL)
    // between 0 and n random particles
    {
      _P.clear();
      std::cerr << "Element::InitRandom" << std::endl;
      abort();
    }
    void AddRandom(double vx, double vy, // navier-stokes velocity
		   int n,
		   double minradius, double maxradius,
		   double minheight, double maxheight,
		   const std::vector<BoundaryLine> &BL)
    // between 0 and n random particles
    {
      std::random_device rd;
      std::mt19937 mt(rd());
      std::uniform_real_distribution<double> posH(0, H);
      
      std::uniform_real_distribution<double> Pradius(minradius, maxradius);
      std::uniform_real_distribution<double> Pheight(minheight, maxheight);
      for (int i = 0; i < rd() % n; ++i)
	{

	  // Particle(_x, _y, _vx, _vy, _height, _radius)
	  Particle p(X0 + posH(mt), Y0 + posH(mt), 0.0, 0.0,
		     Pheight(mt), Pradius(mt));
	  
	  // check that they are within the mesh
	  if (p.x < 0)
	    continue;
	  if (p.y < 0)
	    continue;
	  if (p.y > 2.0)
	    continue;

	  // to generate random size and concavity
	  Zufallszahlengenerator Z;
	  Z.initsimple();

	  // partical maximal velocity at the beginning
	  p.vx = vx;
	  p.vy = vy;

	  if (boundary_lines_index.size() == 0) // check if particle inside domain
	    _P.push_back(p);
	  else
	    {
	      bool inside = true;
	      for (const auto &blit : boundary_lines_index)
		if (BL[blit].distance(p) > 0)
		  inside = false;
	      if (inside)
		_P.push_back(p);
	    }

	
	}
    }

    ////////////////////////////////////////////////// Motion of particles
    // advection without any checking
    // e_vx/e_vy is velocity in the element nodes
    inline void advect(const std::array<double, 4> &e_vx,
		       const std::array<double, 4> &e_vy,
		       const double dt)
    {
      // average velocity in element
      double vxavg = 0.25 * (e_vx[0] + e_vx[1] + e_vx[2] + e_vx[3]);
      double vyavg = 0.25 * (e_vy[0] + e_vy[1] + e_vy[2] + e_vy[3]);
      assert(finite(vxavg));
      assert(finite(vyavg));

      // rotational velocity by stokes theorem
      // 1/2 curl u
      double curl_v = 0.5 * 0.5 / H * ((e_vx[0] + e_vx[1]) + (e_vy[1] + e_vy[3]) - (e_vx[2] + e_vx[3]) - (e_vy[2] + e_vy[0]));
      assert(finite(curl_v));

      // compute rotational velocity
      double NSomega = 0;
      double tx[4] = {1.0, 1.0, -1.0, -1.0};
      double ty[4] = {-1.0, 1.0, 1.0, -1.0};
      for (int i = 0; i < 4; ++i)
	NSomega += 0.25 / (sqrt(2.0) / 2.0 * H) * ((e_vx[i] - vxavg) * tx[i] + (e_vy[i] - vyavg) * ty[i]);
    assert(finite(NSomega));

      for (int i = 0; i < _P.size(); ++i)
	{
	  double X = (_P[i].x - X0) / H;
	  double Y = (_P[i].y - Y0) / H;
    

	  double vx = e_vx[0] * (1. - X) * (1. - Y) + e_vx[1] * X * (1. - Y) + e_vx[2] * (1. - X) * Y + e_vx[3] * X * Y;
	  double vy = e_vy[0] * (1. - X) * (1. - Y) + e_vy[1] * X * (1. - Y) + e_vy[2] * (1. - X) * Y + e_vy[3] * X * Y;
      assert(finite(vx));
      assert(finite(vy));
      
	  _P[i].advect(vx, vy, curl_v, dt);

	}
    }

    inline void collide_two_particles(Particle &P1, Particle &P2, double dt)
    {
      
      
      double diff_x = sqrt((P1.x - P2.x)*(P1.x - P2.x));
      double diff_y = sqrt((P1.y - P2.y)*(P1.y - P2.y));
      double dist = sqrt(diff_x*diff_x + diff_y*diff_y);


      if (dist < P1.radius+P2.radius && P1.x < 7*0.016 && P1.x > 8*0.016 && P1.y > 2*0.016 && P2.x < 7*0.016 && P2.x > 8*0.016 && P2.y > 2*0.016)
      //if (dist < P1.radius+P2.radius)
	{   
    
      P1.x -= P1.vx*dt;
      P1.y -= P1.vy*dt;

      P2.x -= P2.vx*dt;
      P2.y -= P2.vy*dt;
      
        
	  // relative unit-position
	  double nx = P1.x - P2.x;
	  double ny = P1.y - P2.y;

      assert(finite(nx));
      assert(finite(ny));

	  double nn = sqrt(nx * nx + ny * ny);
      
	  nx /= nn;
	  ny /= nn;

	  // relative velocity
	  double vx = P1.vx - P2.vx;
	  double vy = P1.vy - P2.vy;
      
	  // check if particles get closer or if they move away
	  double relvel = nx * vx + ny * vy;

      assert(finite(relvel));
	  if (relvel > 0) // moving away from each other - no contact
	    return;

	  // compute reduced mass
	  double rm = P1.mass() * P2.mass() / (P1.mass() + P2.mass());
      
	  // restitution coefficent
    double k = 0.9; // 1 = elastic, 0 = inealstic collision
	  double rc = 1.0 + k; 
	  
	  // change of momentum in normal direction
	  double deltaP = -rm * rc * relvel; // relvel = v * n is relative normal velocity
	  assert(finite(deltaP));

	  // update velocity
	  P1.vx += deltaP * nx / P1.mass();
	  P1.vy += deltaP * ny / P1.mass();

	  P2.vx -= deltaP * nx / P2.mass();
	  P2.vy -= deltaP * ny / P2.mass();

      assert(finite(P1.vx));
      assert(finite(P2.vx));
      assert(finite(P1.vy));
      assert(finite(P2.vy));

    P1.x += P1.vx*dt;
    P1.y += P1.vy*dt;
    
    P2.x += P2.vx*dt;
    P2.y += P2.vy*dt;
      
	} // if collision
 
    }


    inline void collide_particles(double dt)
    {
      if (_P.size() < 2)
	return;
        
      for (int i = 0; i < _P.size() - 1; ++i)
	for (int j = i + 1; j < _P.size(); ++j){
        
      collide_two_particles(_P[i], _P[j], dt);
  
      
    } // end for loop
    }
    inline void collide_particles(Element &E, double dt)
    {
      if (_P.size() == 0 || E._P.size() == 0){
	      return;
      }

    for (int i = 0; i < _P.size(); ++i){ 
	for (int j = 0; j < E._P.size(); ++j){
  
      collide_two_particles(_P[i], E._P[j], dt);
   
    } }
    
    }
  

    inline void collide_boundary(const std::vector<BoundaryLine> &BL,
				 const std::vector<BoundaryCircle> &BC)
    {

      for (const auto it : boundary_lines_index) // lines
	for (int i = 0; i < _P.size(); ++i)
    {
       
	  if (_P[i].radius > 0) // skip deleted ones
	    {
	     
	      double d = BL[it].distance(_P[i]);

	      if (d > 0) // outside of the boundary
		{
		 
		  BL[it].flip(d, _P[i]);
		}
	    }
    }
      for (const auto it : boundary_circles_index) // circles
	for (int i = 0; i < _P.size(); ++i)
	  if (_P[i].radius > 0) // skip deleted
	    {
        
	      // check if particle is beyond the boundary
	      double d = BC[it].distance(_P[i]);

	      if (d >= 0) // outside of the boundary
		{
		  if (BC[it].color == 0) // outflow boundary
		    {
		      _P[i].radius = -1; // kill particle
		    }
		  else

		    // for new domain
		    if ((BC[it].color != 8) & (BC[it].color != 9)) // no flip at inflow
		      {
			BC[it].flip(d, _P[i]);
		      }
		}
	    }
    }

    void print_particles(std::ofstream &OUT) const
    {
      for (const auto &p : _P)
	p.print(OUT);
    }
  };

}

/*----------------------------   element.h     ---------------------------*/
/* end of #ifndef __elementH */
#endif
/*----------------------------   element.h     ---------------------------*/
