/*----------------------------   gascoignetransfer.h
 * ---------------------------*/
/*      $Id:$                 */
#ifndef __gascoignetransfer_H
#define __gascoignetransfer_H
/*----------------------------   gascoignetransfer.h
 * ---------------------------*/

#include "lattice.h"
#include "baseq22d.h"
#include "finiteelement.h"
#include "gascoigne.h"
#include "gascoignemesh2d.h"
#include "transformation2d.h"


namespace GasParticle
{


  class VelocityTransfer
  { 
    
    std::vector <int> _gas_id_2_lattice_id;
    std::vector <int> _gas_cell_2_lattice;
    std::vector <int> lattice2gascoigne;
    std::vector<int>
        _lc_to_patch; // lattice coordinate index (iy*(NX+1)+ix) to patch
    std::vector<Gascoigne::Vertex2d>
        _lc_to_refpoint; // stores point in reference patch
    std::vector<std::vector<int>>
        _nodes_in_patch; // for every patch, stores the lattice nodes that are
                         // updated here




  public:

    // //// Access
    // void set_velocity_in_element(Lattice& L,
    // 				 std::array<double,4> vx,
    // 				 std::array<double,4> vy,
    // 				 int i, int NX)
    // {
    //   vx[0]=_VX[i];      vy[0]=_VY[i];
    //   vx[1]=_VX[i+1];    vy[1]=_VY[i+1];
    //   vx[2]=_VX[i+NX];   vy[2]=_VY[i+NX];
    //   vx[3]=_VX[i+NX+1]; vy[3]=_VY[i+NX+1];
    // }

    // Initializes the Gascoigne-to-Lattice transfer
    // by determining for every lattice-coordinate:
    // - the patch of the Gascoigne mesh
    // - the coordinate in the reference element


    
    void init_seaice_velocity_to_lattice(Lattice &L,
                                  const Gascoigne::GascoigneMesh2d &M);

    void init_velocity_to_lattice(Lattice &L,
                                  const Gascoigne::GascoigneMesh2d &M);


    // transfers the velocity (in components 1 & 2 of u) to the lattice
    void velocity_to_lattice(Lattice &L,
                             const Gascoigne::GlobalVector &u,
                             const Gascoigne::GascoigneMesh2d &M);

    // transfers the velocity (in components 1 & 2 of u) to the lattice                                                                                                                                               
    void velocity_seaice_to_lattice(Lattice &L,
                             const Gascoigne::GlobalVector &u,
                             const Gascoigne::GascoigneMesh2d &M) const;

    void seaice_tracers_to_cell(Lattice &L,
				 Gascoigne::GlobalVector& u,
				const Gascoigne::GascoigneMesh2d& M) const;


    // .................................................. //  some tools

    // computes the bounding box of a patch given by the dof's in iop
    void bounding_box_of_patch(Gascoigne::Vertex2d &bmin,
                               Gascoigne::Vertex2d &bmax,
                               const Gascoigne::IntVector &iop,
                               const Gascoigne::GascoigneMesh2d &M);

    // transformation matrix for one finite element patch
    void init_transformation(Gascoigne::nmatrix<double> &T,
                             const Gascoigne::IntVector &iop,
                             const Gascoigne::GascoigneMesh2d &M) const;

    // Newton method for getting the reference coordiante
    void get_reference_coordinate(
        Gascoigne::Transformation2d<Gascoigne::BaseQ22d> &transformation,
        Gascoigne::Vertex2d &vref,
        const Gascoigne::Vertex2d &v) const;

    // corresponds to universal_point, initializes finite element function
    void get_velocity(const std::array<double, 9> &UX,
                      const std::array<double, 9> &UY,
                      Gascoigne::FiniteElement<
                          2,
                          1,
                          Gascoigne::Transformation2d<Gascoigne::BaseQ22d>,
                          Gascoigne::BaseQ22d> &finiteelement,
                      double &VX,
                      double &VY);

  };

} // namespace GasParticle


/*----------------------------   gascoignetransfer.h
 * ---------------------------*/
/* end of #ifndef __gascoignetransfer_H */
#endif
/*----------------------------   gascoignetransfer.h
 * ---------------------------*/
