#include "loop.h"
#include "gascoignemesh2d.h"
#include <time.h>
#include  "backup.h"
#include  "stdmultilevelsolver.h"
#include  "sparseblockmatrix.h"
#include  "fmatrixblock.h"
#include  "compose_name.h"
#include <algorithm>
#include <iostream>

#include "../lib/gascoignetransfer.h"

using namespace std;

#include "stopwatch.h"

extern ofstream ELLIPSE_OUT;

double TIME, DT, DTSUB, DELTAMIN,DTUL,starttime,dt_print;
extern int zahl;
extern double STEUERUNG_MU;

namespace Gascoigne
{
  extern Timer GlobalTimer;
  
  class VertexCompare
  {
  public:
    bool operator() (const Vertex2d& v1, const Vertex2d& v2) const
    {
      if (v1.x()<v2.x()-1.e-12)
	return true;
      if (v1.x()>v2.x()+1.e-12)
	return false;

      if (v1.y()<v2.y()-1.e-12)
	return true;

      return false;
    }
  };
  
  void Loop::FVStep(const vector<vector<int> > &FV,
		    const vector<vector<Vertex2d> > &FV_midpoint,
		    GlobalVector& H,
		    const GlobalVector& V,
		    double dtFV, int M)
  {
    double ih = M/0.512; // inverse of the mesh size!!! 
    
    GlobalVector HOLD = H;
    for (int iy=0;iy<M;++iy)
      for (int ix=0;ix<M;++ix)
	{
	  int q = FV[ix][iy];

	  const IntVector& ioc = GetMultiLevelSolver()->GetSolver()->GetMesh()->IndicesOfCell(q);
	  
	  for (int c=0;c<2;++c)
	    {
	      H(q,c)=HOLD(q,c);

	      //  2  3
	      //  0  1

	      
	      if (ix>0) // to the left
		{
		  double vx = .5*V(ioc[0],0) + .5*V(ioc[2],0);
		  double vy = .5*V(ioc[0],1) + .5*V(ioc[2],1);

		  int ql = FV[ix-1][iy];
		  double flux = HOLD(q,c)*(-1.0)*vx;
		  if (-1.0*vx<0)
		    flux = HOLD(ql,c)*(-1.0)*vx;
		  
		  H(q,c) -= flux*ih/0.5*dtFV;
		}
	      
	      if (ix<M-1) // to the right
		{
		  double vx = .5*V(ioc[3],0) + .5*V(ioc[1],0);
		  double vy =  .5*V(ioc[3],1) + .5*V(ioc[1],1);

		  int qr = FV[ix+1][iy];
		  double flux = HOLD(q,c)*(1.0)*vx;
		  if (+1.0*vx<0)
		    flux = HOLD(qr,c)*(1.0)*vx;
		  H(q,c) -= flux*ih/0.5*dtFV;
		}

	      if (iy>0) // bottom
		{
		  double vx = .5*V(ioc[0],0) + .5*V(ioc[1],0);
		  double vy = .5*V(ioc[0],1) + .5*V(ioc[1],1);

		  int qu = FV[ix][iy-1];
		  double flux = HOLD(q,c)*(-1.0)*vy;
		  if (-1.0*vy<0)
		    flux = HOLD(qu,c)*(-1.0)*vy;
		  
		  H(q,c) -= flux*ih/0.5*dtFV;
		}
	      if (iy<M-1) // top
		{
		  double vx = .5*V(ioc[2],0) + .5*V(ioc[3],0);
		  double vy = .5*V(ioc[2],1) + .5*V(ioc[3],1);

		  int qo = FV[ix][iy+1];
		  double flux = HOLD(q,c)*(1.0)*vy;
		  if (+1.0*vy<0)
		    flux = HOLD(qo,c)*(1.0)*vy;
		  H(q,c) -= flux*ih/0.5*dtFV;
		}
	    }
	}
    for (int i=0;i<H.n();++i)
      {
	H(i,1) = min(1.0, H(i,1));
	H(i,1) = max(0.0, H(i,1));
	H(i,0) = max(0.0, H(i,0));
      }

  }
  
  void Loop::run(const std::string& problemlabel)
  {

    GlobalTimer.start("alles");
    GlobalTimer.start("-> Vorbereitung");
      

    zahl=1; 
    double tref; 
    double dtmax, endtime;
    int prerefine;
    string _reloadu,_reloadh,_reloadoldu; 
    if (1)
      {
	DataFormatHandler DFH;
	DFH.insert("dt",   &DT, 0.);
	DFH.insert("dtmax",   &dtmax, 0.);
	DFH.insert("time", &starttime, 0.);
	DFH.insert("endtime", &endtime, 0.);
	DFH.insert("reloadu", &_reloadu); 
	DFH.insert("reloadh", &_reloadh); 
	DFH.insert("reloadoldu", &_reloadoldu); 
	FileScanner FS(DFH);
	FS.NoComplain();

	FS.readfile(_paramfile, "Loop");
	assert(DT>0.0);
	cout<<starttime<<"start"<<endl; }
    if (1)
      {
	DataFormatHandler DFH;
	DFH.insert("prerefine",   &prerefine, 0);
	FileScanner FS(DFH);
	FS.NoComplain();
	FS.readfile(_paramfile, "Mesh");
      }
    if (1)
      {
	DataFormatHandler DFH;
	DFH.insert("deltamin",   &DELTAMIN, 0.);
	DFH.insert("Tref", &tref, 0.);
	FileScanner FS(DFH);
	FS.NoComplain();
	FS.readfile(_paramfile, "Equation");
	assert(DELTAMIN>0.0);
	assert(tref>0.0);
      }
    string discname;
    if (1)
      {
	DataFormatHandler DFH;
	DFH.insert("discname",   &discname);
	FileScanner FS(DFH);
	FS.NoComplain();
	FS.readfile(_paramfile, "Solver");
	assert(discname=="CGQ1" || discname == "CGQ2");
      }

    // Particles                                                                                             
    double latticesize, dt_p, minsize, maxsize, minheight, maxheight;
    int nseed;

    DataFormatHandler DFH;
    FileScanner FS(DFH);
   
    DFH.insert("latticesize", &latticesize, 0.);
    DFH.insert("dt_p", &dt_p, 0.);
    DFH.insert("dt_print", &dt_print, 0.);

    DFH.insert("minsize", &minsize, 0.);
    DFH.insert("maxsize", &maxsize, 0.);

    DFH.insert("minheight", &minheight, 0.);
    DFH.insert("maxheight", &maxheight, 0.);

    DFH.insert("particleseed", &nseed, 0.);

    FS.readfile(_paramfile, "Particle");
    assert(latticesize > 0);
    assert(dt_p > 0);
    assert(minsize > 0);
    assert(maxsize >= minsize);

    assert(minheight > 0);
    assert(maxheight >= minheight);

    assert(dt_print <= DT);
    assert(dt_print >= dt_p); 
  
    // vectors for solution and right hand side
    VectorInterface u("u"), f("f"), oldu("oldu"),other("other");
    Matrix A("A");
    
    VectorInterface DGH("DGH","cell"); // mélange
    VectorInterface ICEBERG("ICEBERG", "cell");
    VectorInterface ICE("ICE", "cell");
    
    PrintMeshInformation();
    // initialize problem, solver and vectors 
    GetMultiLevelSolver()->ReInit();
    GetMultiLevelSolver()->SetProblem("seaice");
    GetMultiLevelSolver()->ReInitMatrix(A);
    GetMultiLevelSolver()->ReInitVector(u);
    GetMultiLevelSolver()->ReInitVector(oldu);
    GetMultiLevelSolver()->ReInitVector(f); 
    GetMultiLevelSolver()->ReInitVector(other);
    
    GetMultiLevelSolver()->ReInitVector(DGH);
    GetMultiLevelSolver()->ReInitVector(ICEBERG);
    GetMultiLevelSolver()->ReInitVector(ICE);
    
    GetSolverInfos()->GetNLInfo().control().matrixmustbebuild() = 1;
    GetMultiLevelSolver()->GetSolver()->OutputSettings();

    // number of cells for finite volume and particle lattice
    const GascoigneMesh* M2 = dynamic_cast<const GascoigneMesh2d* >(GetMultiLevelSolver()->GetSolver()->GetMesh());
    assert(M2);

    int ncells = M2->ncells();
    int M = sqrt(ncells);  // number of cells in every direction

    latticesize=0.512/M; // cell size
    std::cout << "FV transport ncells: " << ncells << " M: " << M << std::endl;
    std::cout << "Particle lattice size: "<< latticesize << " maxheight " << maxheight << std::endl;
    assert(ncells = M*M);


    //////////////////////////////////////////////////// Init Lattice
    GasParticle::Lattice L;
    L.BasicInit(_paramfile);
    L.InitFromGascoigneMesh(dynamic_cast<const GascoigneMesh2d *>(GetMultiLevelSolver()->GetSolver()->GetMesh()),
			    latticesize); 
    L.AddBoundaryLine(GasParticle::BoundaryLine(0.0, 0.0, 0.512, 0.0, 1));   // bottom line
    L.AddBoundaryLine(GasParticle::BoundaryLine(0.512, 0.0, 0.512, 0.512, 1));   // right line
    L.AddBoundaryLine(GasParticle::BoundaryLine(0.512, 0.512, 0.0, 0.512, 1));   // top line 
    L.AddBoundaryLine(GasParticle::BoundaryLine(0., 0.512, 0.0, 0.0, 1));   // left line 
    L.InitBoundaries();

    // create list which assigns a gascoigne cell to element ID
    GasParticle::VelocityTransfer VT;
     VT.init_seaice_velocity_to_lattice(L,
                              dynamic_cast<const GascoigneMesh2d &>(*GetMultiLevelSolver()->GetSolver()->GetMesh()));

    double icearea = 0.512*0.512;
    int Nsqrt = sqrt(nseed*ncells)+1.e-8; // total number of particles = Nsqrt*Nsqrt 
    std::cout << "nseed " << nseed << std::endl;
    std::cout << "ncells " << ncells << " nseed*ncells " << nseed*ncells << std::endl;
    std::cout << "Nsqrt " << Nsqrt << " Nsqrt*Nsqrt " << Nsqrt*Nsqrt << std::endl;

    vector<vector<int > > FV(M,vector<int>(M)); // FV[i][j] is cell number in CR-diskretisation
    vector<vector<Vertex2d > > FV_midpoint(M,vector<Vertex2d>(M)); // FV[i][j] cell midpoint
    map<Vertex2d, int, VertexCompare> MidPointToCellNumber;
    for (int q=0;q<ncells;++q)
      {
	const IntVector &ioc = M2->IndicesOfCell(q);
	Vertex2d v;
	v.equ(0.25,M2->vertex2d(ioc[0]),
	      0.25,M2->vertex2d(ioc[1]),
	      0.25,M2->vertex2d(ioc[2]),
	      0.25,M2->vertex2d(ioc[3]));
	MidPointToCellNumber[v]=q;
      }
    int ix=0;
    int iy=0;
    for (auto it : MidPointToCellNumber)
      {
	FV[ix][iy] = it.second;
	++iy;

	if (iy==M)
	  {
	    iy=0;
	    ++ix;
	  }
    }
    
    for (int iy=0;iy<M;++iy)
      for (int ix=0;ix<M;++ix)
	{
	  const IntVector &ioc = M2->IndicesOfCell(FV[ix][iy]);
	  Vertex2d v;
	  v.equ(0.25,M2->vertex2d(ioc[0]),0.25,M2->vertex2d(ioc[1]),0.25,M2->vertex2d(ioc[2]),0.25,M2->vertex2d(ioc[3]));
	  FV_midpoint[ix][iy] = v;
	}
	

    GlobalVector& glICE = GetMultiLevelSolver()->GetSolver()->GetGV(ICE);
    for (int iy=0;iy<M;++iy)
      for (int ix=0;ix<M;++ix)
	{
	  double x = FV_midpoint[ix][iy].x();
	  double y = FV_midpoint[ix][iy].y();
      
		glICE(FV[ix][iy],0) =  2.0; // thickness
	  	glICE(FV[ix][iy],1) =  0.7; // concentration
      
	}

	
// ICEBERG
//double rparticle = latticesize/2;
double height_in_gridcell = 20.0;
maxheight = height_in_gridcell;// * latticesize * latticesize / M_PI / rparticle / rparticle;

	// equal number of particles in each cell
	double rparticle = latticesize/sqrt(4096)/2; // particle interaction
	double initial_total_area =  nseed*rparticle*rparticle*M_PI*ncells;
    double initial_total_concentration = nseed*(rparticle*rparticle*M_PI/latticesize/latticesize)*ncells; 
	double initial_total_volume = nseed*(rparticle*rparticle*M_PI*maxheight)*ncells; 
	double initial_total_thickness = (nseed*(rparticle*rparticle*M_PI*maxheight)/latticesize/latticesize)*ncells; 
	

	//melange particles interaction 
	for (int ix = 0; ix < 2; ++ix){
        if(ix == 0)
        { 
            GasParticle::Particle p(0.512/2 + 2*rparticle, 0.512/2 + rparticle*0.4, 0.,0.,maxheight,rparticle); 
            L.elements[L.e_id_by_mesh_coord(0.512/2 + 2*rparticle, 0.512/2 + rparticle*0.4)].AddParticle(p);
        } else if (ix == 1) { 
            GasParticle::Particle p(0.512/2-rparticle*1, 0.512/2 - rparticle*0.5, 0.,0.,maxheight,rparticle); 
            L.elements[L.e_id_by_mesh_coord(0.512/2-rparticle*1, 0.512/2 - rparticle*0.5)].AddParticle(p);
    }
    } 

		 
// ######################################################################		     
    
	// Inizialize H and A 
	GlobalVector& glDGH = GetMultiLevelSolver()->GetSolver()->GetGV(DGH);
    GlobalVector& glICEBERG = GetMultiLevelSolver()->GetSolver()->GetGV(ICEBERG);

    VT.seaice_tracers_to_cell( L, glICEBERG,
	 dynamic_cast<const GascoigneMesh2d &>(*GetMultiLevelSolver()->GetSolver()->GetMesh()));
 
    double total_concentration = 0;
	double total_thickness = 0;
	double total_volume = 0;

	  for (int ix=0;ix<ncells;++ix) 
	    {

          total_concentration += glICEBERG(ix,1);
          total_volume += glICEBERG(ix,0);
          
          glICEBERG(ix,0) = glICEBERG(ix,0)/latticesize/latticesize;
		  glICEBERG(ix,1) = glICEBERG(ix,1);
          
		  }


        
		GetMultiLevelSolver()->GetSolver()->CellVisu("Results/dgh",glDGH,0); 
		GetMultiLevelSolver()->GetSolver()->CellVisu("Results/seaice",glICE,0);
        GetMultiLevelSolver()->GetSolver()->CellVisu("Results/iceberg",glICEBERG,0);
        L.write_vtk("Results/p", 0);
 
	
	cout << "------------------------------" << endl; 
	cout << "hybrid ice-mélange" << endl;
	cout << "------------------------------" << endl << endl;
  
	// See-Eis
	GetMultiLevelSolver()->SetProblem("seaice");
	if (_initial=="reload")	
	  {
	    abort();
	    GetMultiLevelSolver()->GetSolver()->Read(u,_reloadu);    
	  }
	else 
	  {
	    InitSolution(u);     
	    GetMultiLevelSolver()->Equ(oldu,1.0,u);
	  }

    GetMultiLevelSolver()->GetSolver()->Visu("Results/v",u,0); 

	double stepback=0.0;
	double writenext= 0;
	int writeiter = 0;
	string res;
  
	nvector<double> functionals;
	
  
	int timeinc=0;
  
	clock_t start, end;
	double cpu_time_used,a;

	TIME=starttime;
 

	nvector<double>      Jtotal,Jtotal1;

	GlobalTimer.stop("-> Vorbereitung");
	for (_iter=1; _iter<=_niter; _iter++)
	  {
	    GlobalTimer.start("-> Iteration");

	    TIME += DT;


	    GetSolverInfos()->GetNLInfo().control().matrixmustbebuild() = 1;

	    cout << "\n time step " << _iter << " " 
		 << "\t" << TIME*tref/60/60/24 << " days" << endl;


	//-----------------------------FV--transport of sea ice thickness H and concentration A	    
	GlobalTimer.start("--> Transport");

	int NSUB = 100;
		
	double dtFV=DT/NSUB;
	for (int ii=1;ii<=NSUB;++ii){
	  FVStep(FV,FV_midpoint,glICE,GetMultiLevelSolver()->GetSolver()->GetGV(u),dtFV,M);
    };
    
     for (int ix=0;ix<ncells;++ix) 
	    {
            
		  glDGH(ix,0) = glICE(ix,0)+ glICEBERG(ix,0); // thickness
		  glDGH(ix,1) = glICE(ix,1)+ glICEBERG(ix,1); // concentration
          
          glDGH(ix,1) = std::min(glDGH(ix,1),1.0);
                        
		  }
		  
    GetMultiLevelSolver()->GetSolver()->CellVisu("Results/dgh",glDGH,_iter);
    GetMultiLevelSolver()->GetSolver()->CellVisu("Results/seaice",glICE,_iter);
    GetMultiLevelSolver()->GetSolver()->CellVisu("Results/iceberg",glICEBERG,_iter);
	    
	GlobalTimer.stop("--> Transport");
	//-----------------------------FV--Transport----------------------------

	    
	    cout<<"Momentum"<<endl;
      	  GlobalTimer.start("--> Momenten"); 
	    GetMultiLevelSolver()->SetProblem("seaice");
	    GetSolverInfos()->GetNLInfo().control().matrixmustbebuild() = 1;
	    GetMultiLevelSolver()->AddNodeVector("oldu", oldu);
	     GetMultiLevelSolver()->GetSolver()->AddCellVector("DGH", DGH);  
		 GetMultiLevelSolver()->GetSolver()->AddCellVector("ICEBERG", ICEBERG);   
	    GetMultiLevelSolver()->GetSolver()->Equ(oldu,1.0, u);
	    GetMultiLevelSolver()->AssembleMatrix(A,u);
	    GetMultiLevelSolver()->ComputeIlu(A,u); 
      
	    res=Solve(A,u,f,"Results/u");
	    assert(res=="converged"); 

	  GetMultiLevelSolver()->GetSolver()->Visu("Results/v",u,_iter); 
	  

	   GetMultiLevelSolver()->DeleteNodeVector("oldu");
	    GetMultiLevelSolver()->GetSolver()->DeleteCellVector("DGH");
		GetMultiLevelSolver()->GetSolver()->DeleteCellVector("ICEBERG"); 
      	    GlobalTimer.stop("--> Momenten"); 

	    //----------------------------- sea ice momentum equation -----------------     
	    //----------------------------- Set up particle advection-----------------     
	    // Transfer velocity to the lattice   
	    GlobalTimer.start("--> velocitytransfer");
	    VT.velocity_seaice_to_lattice(L, GetMultiLevelSolver()->GetSolver()->GetGV(u),
				   dynamic_cast<const GascoigneMesh2d &>(*GetMultiLevelSolver()->GetSolver()->GetMesh()));
	    GlobalTimer.stop("--> velocitytransfer");
 	   
	    // Advect the particles with sub-steps     
	    std::cout << "Advect " << L.nparticles() << " particles in " << DT / dt_p
		      << " sub-steps of size " << dt_p << std::endl;
	    GlobalTimer.start("particles advect");
	    int print_every = dt_print / dt_p;
	    for (int k = 0; k < DT / dt_p; ++k)
	      {
		GlobalTimer.start("particles advect - ad");
		L.advect(dt_p);
		GlobalTimer.stop("particles advect - ad");
		                                                                                    
		GlobalTimer.start("particles advect - t");
		L.transfer();
		GlobalTimer.stop("particles advect - t");
        
        GlobalTimer.start("particles advect - c");
		L.collide_particles(dt_p);
		GlobalTimer.stop("particles advect - c");
        
        GlobalTimer.start("particles advect - t");
		L.transfer();
		GlobalTimer.stop("particles advect - t");
		
        GlobalTimer.start("particles advect - cb");
		L.collide_boundary();
		GlobalTimer.stop("particles advect - cb");
        
        GlobalTimer.start("particles advect - t");
		L.transfer();
		GlobalTimer.stop("particles advect - t");

	      }
	    std::cout << "Particles nach advection" << L.nparticles() << std::endl;

	    GlobalTimer.stop("particles advect");

	    	    GlobalTimer.start("particles i/o");
	      L.write_vtk("Results/p", _iter);
	     GlobalTimer.stop("particles i/o");
	    

	    //-----------------------------FV--Transport---------------------------- 
	    cout<<"average particle to cell"<<endl;
            VT.seaice_tracers_to_cell(L,glICEBERG, dynamic_cast<const GascoigneMesh2d &>(*GetMultiLevelSolver()->GetSolver()->GetMesh()));

	    //----------------------------- Set up particle advection-----------------   

	    writenext+=DT;
      	   

	    GlobalTimer.stop("-> Iteration"); 
	    if (TIME >= endtime*24*60*60/tref) break;
	  
	  }
	
    GlobalTimer.stop("alles");
  
    GlobalTimer.print100();
  }



}






 
