#include "gascoignetransfer.h"
#include <fstream>
#include <iomanip>

using namespace Gascoigne;

namespace GasParticle
{
  void VelocityTransfer::bounding_box_of_patch(Gascoigne::Vertex2d& bmin, Gascoigne::Vertex2d& bmax,
					       const Gascoigne::IntVector& iop,
					       const Gascoigne::GascoigneMesh2d& M)
  {
    bmin = M.vertex2d(iop[0]);
    bmax = bmin;

    for (int i=1;i<iop.size();++i)
      {
	bmin.x() = std::min(bmin.x(), M.vertex2d(iop[i]).x());
	bmin.y() = std::min(bmin.y(), M.vertex2d(iop[i]).y());

	bmax.x() = std::max(bmax.x(), M.vertex2d(iop[i]).x());
	bmax.y() = std::max(bmax.y(), M.vertex2d(iop[i]).y());
      }
  }

  void VelocityTransfer::init_transformation(nmatrix<double>& T, const IntVector& iop,
					     const GascoigneMesh2d& M) const
 {
    for(int ii=0;ii<iop.size();ii++)
      {
	Vertex2d v = M.vertex2d(iop[ii]);
	T(0,ii) = v.x();
	T(1,ii) = v.y();
      }
  }

  void VelocityTransfer::get_reference_coordinate(Transformation2d<BaseQ22d>& transformation, Vertex2d& vref, const Vertex2d& v) const
  {
    // Newton method for finding the reference coordinate
    Vertex2d residual;

    transformation.point(vref);
    residual = transformation.x();

    residual.add(-1.0,v);
    double tol = (1.e-14)*(1.e-14);        // tolerance of newton (squared)
    int newtoniter = 0;

    int MAXITER = 10;
    
    while (residual.x()*residual.x() + residual.y()*residual.y() > tol) // newton loop
      {
	assert(newtoniter<MAXITER);
	++newtoniter;
	transformation.DTI().multtrans(vref,residual,-1.0);
	//transformation.DTI().Mult(vref,residual,-1.0);
	transformation.point(vref);
	residual = transformation.x();
	residual.add(-1.0,v);
      }
  }


  //--------------------------init velo + list von coordinates to lattice node                                                                                                                                                 
  void VelocityTransfer::init_seaice_velocity_to_lattice(Lattice& L,
                                                  const GascoigneMesh2d& M)
  {
    std::cout << "init velocity-to-lattice ... L.NX: " << L.NX << " L.NY: " << L.NY << std::endl;

    // initialize list with coordinates on lattice nodes
     _gas_id_2_lattice_id.resize((L.NX+1)*(L.NY+1));

     for(int i=0; i< ((L.NX+1)*(L.NY+1));++i)
       {
	 const  Vertex2d& v=M.vertex2d(i);
	 int e_ix=int((v.x()-L.X0)/L.H+ 1.e-10);
	 int e_iy=int((v.y()-L.Y0)/L.H+ 1.e-10);
	 
	 assert((v.x()-e_ix *L.H+L.X0)<1.e-10);
	 assert((v.y()-e_iy*L.H+L.Y0)<1.e-10);

	 int e_id=e_iy*(L.NX+1)+e_ix;
	
	 assert(e_id<(L.NX+1)*(L.NY+1));
	 _gas_id_2_lattice_id[i]=e_id; 
       }

     for (int i=0;i<(L.NX+1)*(L.NY+1);++ i)
       assert(_gas_id_2_lattice_id[i]>=0);
   
    // inialize velocity fields and set velocity to zero                                        
    L.VX.resize((L.NX+1)*(L.NY+1));
    for (auto &it : L.VX) it=0;
    L.VY.resize((L.NX+1)*(L.NY+1));

    for (auto &it : L.VY) it=0;
    // initialize liste mit cell_number to lattice
    _gas_cell_2_lattice.resize((L.NX)*(L.NY));

    assert((L.NX)*(L.NY) == M.ncells());
    double RX = 0.512; // domain size
    double RY = 0.512;
        

    for (int i=0;i<M.ncells();++i)
    {
        const IntVector &ioc = M.IndicesOfCell(i);
        Vertex2d v;
	
        v.equ(0.25,M.vertex2d(ioc[0]),0.25,M.vertex2d(ioc[1]),0.25,M.vertex2d(ioc[2]),0.25,M.vertex2d(ioc[3]));
        
        int ix = (v.x()/RX)*L.NX;
        int iy = (v.y()/RY)*L.NY;
        int li = ix + iy * L.NX;

        _gas_cell_2_lattice[i]=li;

      }
      
      lattice2gascoigne.resize(M.ncells(),-1);
      for (int i=0;i<M.ncells();++i)
          lattice2gascoigne[_gas_cell_2_lattice[i]] = i;
      for (int i=0;i<M.ncells();++i)
            assert(lattice2gascoigne[i]>-1);
  }


  void VelocityTransfer::velocity_seaice_to_lattice(Lattice& L,
						    const GlobalVector& u,
						    const GascoigneMesh2d& M) const
    
  {
    for(int i=0; i<(L.NX+1)*(L.NY+1);++i)
      {  
	L.VX[_gas_id_2_lattice_id[i]]= u(i,0);
	L.VY[_gas_id_2_lattice_id[i]]= u(i,1);
	
    assert(finite(u(i,0)));
      assert(finite(u(i,1)));
      }
  }

  void VelocityTransfer::seaice_tracers_to_cell(Lattice& L,
						GlobalVector& u,
						const GascoigneMesh2d& M) const
  //vektor with two components with average thickness and concentration
  {
        assert(L.NX*L.NY == u.n());
        assert(_gas_cell_2_lattice.size() == u.n());


for(int i=0; i<u.n();++i)
      {
	u(i,1)=0.0;
	u(i,0)=0.0;
      }
	
     
     int i = 0;
     for (int iy=0; iy<L.NY;++iy)
       for (int ix=0; ix<L.NX;++ix,++i)
       {
          
          std::array<double,9> element_concentration({0,0,0,0,0,0,0,0,0});
          std::array<double,9> element_volume({0,0,0,0,0,0,0,0,0});
          
       
          element_concentration = L.elements[i].iceconcentration();
          element_volume     = L.elements[i].icevolume();

          u(lattice2gascoigne[i],1)+=element_concentration[0];
          u(lattice2gascoigne[i],0)+=element_volume[0];
         
		 /**
		  * 1 2 3
		  * 4 0 5
		  * 6 7 8 
		  **/
          
          
          // left
	  if (ix>0)
	    {
	      u(lattice2gascoigne[i-1],1)+=element_concentration[4];
	      u(lattice2gascoigne[i-1],0)+=element_volume[4];
         
	    }
	  else
	    {
	      u(lattice2gascoigne[i],1)+=element_concentration[4];
	      u(lattice2gascoigne[i],0)+=element_volume[4];
        
	    }
	  
	  // right
	  if (ix<L.NX-1)
	    {
	      u(lattice2gascoigne[i+1],1)+=element_concentration[5];  
	      u(lattice2gascoigne[i+1],0)+=element_volume[5];
         
	    }
	  else
	    {
	      u(lattice2gascoigne[i],1)+=element_concentration[5];
	      u(lattice2gascoigne[i],0)+=element_volume[5];
          
	    }
	      
	  // bottom
	  if (iy>0)
	    {
	      u(lattice2gascoigne[i-L.NX],1)+=element_concentration[7];  
	      u(lattice2gascoigne[i-L.NX],0)+=element_volume[7];
         
	    }
	  else
	    {
	      u(lattice2gascoigne[i],1)+=element_concentration[7];
	      u(lattice2gascoigne[i],0)+=element_volume[7];
          
	    }
	  // top
	  if (iy<L.NY-1)
	    {
	      u(lattice2gascoigne[i+L.NX],1)+=element_concentration[2]; 
	      u(lattice2gascoigne[i+L.NX],0)+=element_volume[2];
          
	    }
	  else
	    {
	        u(lattice2gascoigne[i],1)+=element_concentration[2];
	        u(lattice2gascoigne[i],0)+=element_volume[2];
            
	    }
	    
        // left down
	  if ((ix>0) && (iy>0) )
	    {
	      u(lattice2gascoigne[i-L.NX-1],1)+=element_concentration[6];  
              u(lattice2gascoigne[i-L.NX-1],0)+=element_volume[6];
              
	    }
	  else
	    {
	      if (ix>0) // bottom but not left -> to the left
		{
		  u(lattice2gascoigne[i-1],1)+=element_concentration[6]; 
		  u(lattice2gascoigne[i-1],0)+=element_volume[6];
          
		}
	      else if (iy>0) // left but not bottom -> to bottom
		{
		  u(lattice2gascoigne[i-L.NX],1)+=element_concentration[6]; 
		  u(lattice2gascoigne[i-L.NX],0)+=element_volume[6];
          
		}	
	      else // left bottom -> middle
		{
		  u(lattice2gascoigne[i],1)+=element_concentration[6]; 
		  u(lattice2gascoigne[i],0)+=element_volume[6];
          
		}
	    }
	    
	    // right bottom
	  if ((ix<L.NX-1) && (iy>0) )
	    {
	      u(lattice2gascoigne[i-L.NX+1],1)+=element_concentration[8];  
              u(lattice2gascoigne[i-L.NX+1],0)+=element_volume[8];
              
	    }
	  else
	    {
	      if (ix<L.NX-1) //bottom but not right -> right 
		{
		  u(lattice2gascoigne[i+1],1)+=element_concentration[8]; 
		  u(lattice2gascoigne[i+1],0)+=element_volume[8];
          
		}
	      else if (iy>0) // right but not bottom -> bottom 
		{
		  u(lattice2gascoigne[i-L.NX],1)+=element_concentration[8]; 
		  u(lattice2gascoigne[i-L.NX],0)+=element_volume[8];
          
		}	
	      else // right bottom -> middle
		{
		  u(lattice2gascoigne[i],1)+=element_concentration[8]; 
		  u(lattice2gascoigne[i],0)+=element_volume[8];
         
		}
	    }
	    
	     // upper left 
	  if ((ix>0) && (iy<L.NY-1) )
	    {
	      u(lattice2gascoigne[i+L.NX-1],1)+=element_concentration[1];  
              u(lattice2gascoigne[i+L.NX-1],0)+=element_volume[1];
              
	    }
	  else
	    {
	      if (ix>0) // up but not left -> left
		{
		  u(lattice2gascoigne[i-1],1)+=element_concentration[1]; 
		  u(lattice2gascoigne[i-1],0)+=element_volume[1];
          
		}
	      else if (iy<L.NY-1) // left but not up -> up
		{
		  u(lattice2gascoigne[i+L.NX],1)+=element_concentration[1]; 
		  u(lattice2gascoigne[i+L.NX],0)+=element_volume[1]; 
		}	
	      else // upper left -> middle
		{
		  u(lattice2gascoigne[i],1)+=element_concentration[1]; 
		  u(lattice2gascoigne[i],0)+=element_volume[1];
        
		}
	    }
	    
	    // upper right
	  if ((ix<L.NX-1) && (iy<L.NY-1) )
	    {
	      u(lattice2gascoigne[i+L.NX+1],1)+=element_concentration[3]; 
              u(lattice2gascoigne[i+L.NX+1],0)+=element_volume[3];
             
	    }
	  else
	    {
	      if (ix<L.NX-1) // top but not right -> right
		{
		  u(lattice2gascoigne[i+1],1)+=element_concentration[3]; 
		  u(lattice2gascoigne[i+1],0)+=element_volume[3];
         
		}
	      else if (iy< L.NY-1) // right but not top -> up 
		{
		  u(lattice2gascoigne[i+L.NX],1)+=element_concentration[3]; 
		  u(lattice2gascoigne[i+L.NX],0)+=element_volume[3];
         
		}	
	      else // upper right -> middle
		{
		  u(lattice2gascoigne[i],1)+=element_concentration[3]; 
		  u(lattice2gascoigne[i],0)+=element_volume[3];
         
		}
	    }  
        
        
            
          
       }

    for(int i=0; i<u.n();++i)
    {
        
		assert(u(i,0)>=0);
		assert(u(i,1)>=0);
		if(u(i,1)==0)
		{
			assert(u(i,0)==0);
		}
		u(i,1) = std::min(u(i,1),1.0);
	}

  }


  ////////////////////////////////////////////////// INIT
  void VelocityTransfer::init_velocity_to_lattice(Lattice& L,
						  const GascoigneMesh2d& M)
  {
    std::cout << "init velocity-to-lattice ... " << std::endl;

    // initialize lattice-node-indextopatch
    _lc_to_patch.resize((L.NX+1)*(L.NY+1));
    for (auto &it : _lc_to_patch)
      it=-1;
    // initialize lattice-node to reference point
    _lc_to_refpoint.resize((L.NX+1)*(L.NY+1));


    // inialize velocity fields and set velocity to zero
    L.VX.resize((L.NX+1)*(L.NY+1));
    for (auto &it : L.VX) it=0;
    L.VY.resize((L.NX+1)*(L.NY+1));
    for (auto &it : L.VY) it=0;


    // Loop over all patches of the mesh. Only makes sense for Q2-Discretizations!
    for (int patch=0;patch<M.npatches();++patch)
      {
	const IntVector& iop = *M.IndicesOfPatch(patch);  // dof-indices of the mesh


	Vertex2d bmin,bmax;                      // compute the bounding box of the patch
	bounding_box_of_patch(bmin,bmax,iop,M);
	int ixmin,ixmax,iymin,iymax;             // compute the indices of the lattice
	L.e_id_to_e_coord(L.e_id_by_mesh_coord(bmin.x(),bmin.y()),ixmin,iymin);
	L.e_id_to_e_coord(L.e_id_by_mesh_coord(bmax.x(),bmax.y()),ixmax,iymax);
	std::cout << L.e_id_by_mesh_coord(bmin.x(),bmin.y()) << "\t" << L.e_id_by_mesh_coord(bmax.x(),bmax.y()) << std::endl;
	std::cout << bmin<< "\t \t "<< bmax << std::endl;
	std::cout << "init velocity-to-lattice ... " << ixmax << " " << ixmin << std::endl;

	assert(ixmax>=ixmin);
	assert(iymax>=iymin);

	// Initialize a Finite Element of this patch
	nmatrix<double> transmatrix;
	transmatrix.memory(2,9);
	init_transformation(transmatrix, iop, M);
	Transformation2d<BaseQ22d> transformation;
	transformation.ReInit(transmatrix);

	for (int iy=iymin+1;iy<=iymax;++iy) // run over all elements within the bounding box
	  for (int ix=ixmin+1;ix<=ixmax;++ix)
	    {
	      int lc_i = iy*(L.NX+1)+ix; // index of the lattice node
	      assert(lc_i<_lc_to_patch.size());

	      Vertex2d v;
	      L.mesh_coord_by_e_coord(v.x(),v.y(), ix,iy); // transform the lattice node to the coordinate in the mesh
	      Vertex2d vref(0.5,0.5);                                    // get the corresponding coordinate in the reference patch

	      get_reference_coordinate(transformation,vref,v);

	      // is this node already set?
	      if (_lc_to_patch[lc_i]>=0)
		{
		  // node already set. Only overwrite patch and reference coordiante if within the patch
		  if ( (vref.x()>0.0-1.e-12)&&(vref.x()<1.0+1.e-12)&&(vref.y()>0.0-1.e-12)&&(vref.y()<1.0+1.e-12))
		    {
		      _lc_to_patch[lc_i] = patch;
		      _lc_to_refpoint[lc_i] = vref;
		    }
		}
	      else  // new node. Already set patch and reference coordinate, even if outside patch
		{
		  _lc_to_patch[lc_i] = patch;
		  _lc_to_refpoint[lc_i] = vref;
		}
	    }
      }
    for (int i=0;i<_lc_to_patch.size();++i)
      {
	if (_lc_to_patch[i]==-1)
	  {
	    double x,y;
	    int iy = i/(L.NX+1);
	    int ix = i%(L.NX+1);
	    L.mesh_coord_by_e_coord(x,y,ix,iy);
		
	  }
      }

    // initialize list storing which nodes belong to each patch
    _nodes_in_patch.resize(M.npatches());
    for (auto &it : _nodes_in_patch)
      it.clear();

    int ni=0;
    for (int i=0;i<_lc_to_patch.size();++i)
      {
	int patch = _lc_to_patch[i];
	if (patch==-1) continue;     // patch never set
	++ni;

	assert(_lc_to_patch[i]<M.npatches());
	_nodes_in_patch[_lc_to_patch[i]].push_back(i);
      }
    std::cout << ni << " / "
	      << (L.NX+1)*(L.NY+1) << " lattice nodes found" << std::endl;

  }


  // LocalToGlobal
  void VelocityTransfer::get_velocity(const std::array<double,9>& UX,
				      const std::array<double,9>& UY,
				      FiniteElement<2,1,Transformation2d<BaseQ22d>, BaseQ22d>& finiteelement,
				      double& VX,
				      double& VY)
  {
    TestFunction phi;
    VX=0.0;
    VY=0.0;

    assert(finiteelement.n()==9);

    for (int i=0; i<finiteelement.n(); i++) // run over all basis functions
    {
      finiteelement.init_test_functions(phi,1.,i);
      VX += UX[i]*phi.m();
      VY += UY[i]*phi.m();
    }
  }


  ////////////////////////////////////////////////// Transfer
  void VelocityTransfer::velocity_to_lattice(Lattice& L,
					     const GlobalVector& u,
					     const GascoigneMesh2d& M)
  {
    for (auto &it : L.VX) it = 0;
    for (auto &it : L.VY) it = 0;

    // Loop over all patches of the mesh. Only makes sense for Q2-Discretizations!
#pragma omp parallel for
    for (int patch=0;patch<M.npatches();++patch)
      {
	if (_nodes_in_patch.size()==0)
	  continue;

	const IntVector& iop = *M.IndicesOfPatch(patch);  // dof-indices of the mesh
	assert(iop.size()==9);

	// Initialize a Finite Element of this patch
	nmatrix<double> transmatrix;
	transmatrix.memory(2,9);
	init_transformation(transmatrix, iop, M);

	FiniteElement<2,1,Transformation2d<BaseQ22d>, BaseQ22d> finiteelement;
	finiteelement.ReInit(transmatrix);

	// Get velocity LocalVector
	std::array<double,9> UX;
	std::array<double,9> UY;
	for (int i=0;i<9;++i)
	  {
	    UX[i] = u(iop[i],0);
	    UY[i] = u(iop[i],1);
	  }

	for (auto lc_i : _nodes_in_patch[patch])
	  {
	    finiteelement.point(_lc_to_refpoint[lc_i]); // initialize the finite element in the ref point

	    get_velocity(UX,UY,finiteelement,L.VX[lc_i],L.VY[lc_i]);

	  }
      }


    return;
    // used to transform between the local numering of the nodes
    // in one element to the global numbering in the lattice
    int DIRX[4] = {0,1,1,0};
    int DIRY[4] = {0,0,1,1};


    for (int e=0;e<L.elements.size();++e)
      {
	for (const auto &bl : L.boundary_lines)
	  {
	    double distance[4] = // positive is outside
	      {bl.raw_distance(L.elements[e].X0,L.elements[e].Y0),
	       bl.raw_distance(L.elements[e].X1,L.elements[e].Y0),
	       bl.raw_distance(L.elements[e].X1,L.elements[e].Y1),
	       bl.raw_distance(L.elements[e].X0,L.elements[e].Y1)};
	    int npos=0;
	    int nneg=0;
	    for (int i=0;i<4;++i)
	      {
		if (distance[i]>0) ++npos;
		if (distance[i]<0) ++nneg;
	      }
	    if ( (npos==4) || (nneg==4)) continue;

	    for (int edge=0;edge<4;++edge)
	      {
		if (distance[edge]*distance[(edge+1)%4]<0) // edge is cut
		  {
		    // which point of the edge is inside the domain, which is outside?
		    int insidepoint  = distance[edge]<0?edge:(edge+1)%4;
		    int outsidepoint = distance[edge]<0?(edge+1)%4:edge;

		    // coordinates of the element
		    int ex,ey;
		    L.e_id_to_e_coord(e,ex,ey);

		    // id of the node in the lower left of the element
		    int n_id = ey*(L.NX+1)+ex;
		    // id of the nodes of insidepoint and outsidepoint
		    int inside_id  = n_id + DIRX[insidepoint]  + (L.NX+1)*DIRY[insidepoint];
		    int outside_id = n_id + DIRX[outsidepoint] + (L.NX+1)*DIRY[outsidepoint];


		    int color = bl.color;
		    if ((color==0) || (color == 8)) // outflow and inflow
		      {
			L.VX[outside_id] = L.VX[inside_id];
			L.VY[outside_id] = L.VY[inside_id];
		      }
		    else // wall with Dirichlet 0
		      {
			double xi,yi,xo,yo; // coordinates of inside and outside nodes
			L.mesh_coord_by_e_coord(xi,yi,ex+DIRX[insidepoint], ey+DIRY[insidepoint]);
			L.mesh_coord_by_e_coord(xo,yo,ex+DIRX[outsidepoint],ey+DIRY[outsidepoint]);
			double s = bl.position_between_points(xi,yi, xo,yo);
			L.VX[outside_id] = L.VX[inside_id]*(1.0-1.0/s);
			L.VY[outside_id] = L.VY[inside_id]*(1.0-1.0/s);
		      }
		  }
	      }
	  }
      }


    for (int e=0;e<L.elements.size();++e)
      {
	for (const auto bl : L.boundary_circles)
	  {
	    double distance[4] = // positive is outside
	      {bl.raw_distance(L.elements[e].X0,L.elements[e].Y0),
	       bl.raw_distance(L.elements[e].X1,L.elements[e].Y0),
	       bl.raw_distance(L.elements[e].X1,L.elements[e].Y1),
	       bl.raw_distance(L.elements[e].X0,L.elements[e].Y1)};
	    int npos=0;
	    int nneg=0;
	    for (int i=0;i<4;++i)
	      {
		if (distance[i]>0) ++npos;
		if (distance[i]<0) ++nneg;
	      }
	    if ( (npos==4) || (nneg==4)) continue;

	    for (int edge=0;edge<4;++edge)
	      {
		if (distance[edge]*distance[(edge+1)%4]<0) // edge is cut
		  {
		    // which point of the edge is inside the domain, which is outside?
		    int insidepoint  = distance[edge]<0?edge:(edge+1)%4;
		    int outsidepoint = distance[edge]<0?(edge+1)%4:edge;

		    // coordinates of the element
		    int ex,ey;
		    L.e_id_to_e_coord(e,ex,ey);

		    // id of the node in the lower left of the element
		    int n_id = ey*(L.NX+1)+ex;
		    // id of the nodes of insidepoint and outsidepoint
		    int inside_id  = n_id + DIRX[insidepoint]  + (L.NX+1)*DIRY[insidepoint];
		    int outside_id = n_id + DIRX[outsidepoint] + (L.NX+1)*DIRY[outsidepoint];


		    int color = bl.color;
		    if ((color==0) || (color == 8)) // outflow and inflow
		      {
			L.VX[outside_id] = L.VX[inside_id];
			L.VY[outside_id] = L.VY[inside_id];
		      }
		    else // wall with Dirichlet 0
		      {
			double xi,yi,xo,yo; // coordinates of inside and outside nodes
			L.mesh_coord_by_e_coord(xi,yi,ex+DIRX[insidepoint], ey+DIRY[insidepoint]);
			L.mesh_coord_by_e_coord(xo,yo,ex+DIRX[outsidepoint],ey+DIRY[outsidepoint]);
			double s = bl.position_between_points(xi,yi, xo,yo);
			L.VX[outside_id] = L.VX[inside_id]*(1.0-1.0/s);
			L.VY[outside_id] = L.VY[inside_id]*(1.0-1.0/s);
		      }
		  }
	      }
	  }
      }




  }



}
